///////////////////////////////////////////////////////////////////////////////
// File: times.cpp                                                           //
// Description: example program that computes execution times                //
// This file is part of the SISCone project.                                 //
// For more details, see http://www.lpthe.jussieu.fr/~soyez/siscone          //
//                                                                           //
// Copyright (c) 2006 Gavin Salam and Gregory Soyez                          //
//                                                                           //
// This program is free software; you can redistribute it and/or modify      //
// it under the terms of the GNU General Public License as published by      //
// the Free Software Foundation; either version 2 of the License, or         //
// (at your option) any later version.                                       //
//                                                                           //
// This program is distributed in the hope that it will be useful,           //
// but WITHOUT ANY WARRANTY; without even the implied warranty of            //
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             //
// GNU General Public License for more details.                              //
//                                                                           //
// You should have received a copy of the GNU General Public License         //
// along with this program; if not, write to the Free Software               //
// Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA //
//                                                                           //
// $Revision:: 128                                                          $//
// $Date:: 2007-01-18 20:07:57 +0100 (Thu, 18 Jan 2007)                     $//
///////////////////////////////////////////////////////////////////////////////

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <sys/time.h>
#include <iostream>
#include <math.h>

#include "ranlux.h"
#include "momentum.h"
#include "protocones.h"
#include "split_merge.h"

#define Nruns 32
#define R    1.0
#define SMT  0.5

using namespace std;
using namespace siscone;

timeval time_start, time_end;

// compute time spent between time_start and time_end
int time_spent(){
  timeval time_diff;
  
  // compute different with initial time
  time_diff.tv_sec = time_end.tv_sec-time_start.tv_sec;
  if (time_end.tv_usec > time_start.tv_usec){
    time_diff.tv_usec = time_end.tv_usec-time_start.tv_usec;
  } else {
    time_diff.tv_sec--;
    time_diff.tv_usec = (1000000+time_end.tv_usec)-time_start.tv_usec;
  }
  
  return 1000000*time_diff.tv_sec+time_diff.tv_usec;
}



int main(){
  vector<Cmomentum> particles;
  Cstable_cones *engine;
  Csplit_merge *sm;
  double eta,phi;

  // number of events and particles
  int i, nb_proto, N;
  int n_ev, part_inc;

  // number statistics variables
  int nb_total, nb_tested, nb_stables, nb_jets;

  // time statistics variables
  int time_partic, time_alloc, time_analyse, time_splitmerge;

  // save files
  FILE *flux;
  FILE *fnum;
  FILE *fnum2;

  // initialise random number generator
  cout << "initialise random number generator" << endl;
  timeval timestamp;

  gettimeofday(&timestamp, NULL);
  srand(timestamp.tv_usec);

  flux = fopen("times.dat", "w+");
  fnum = fopen("stat.dat", "w+");
  fnum2 = fopen("stat2.dat", "w+");

  N = 1;
  part_inc = 1;
  ranlux_init();
  engine = new Cstable_cones;
  sm = new Csplit_merge;
  do{
    fprintf(stdout, "\r%5d particles\n", N);
    time_partic=time_alloc=time_analyse=time_splitmerge=0;
    nb_total=nb_tested=nb_stables=nb_jets=0;

    for (n_ev=0;n_ev<Nruns;n_ev++){
      // build particle list
      gettimeofday(&time_start, NULL);
      particles.clear();
      for (i=0;i<N;i++){
	eta = -3.0+6.0*rand()/(RAND_MAX+1.0);
	phi = 2.0*M_PI*rand()/(RAND_MAX+1.0);
	particles.push_back(Cmomentum(cos(phi), sin(phi), tanh(eta), 1.0));
      }
      gettimeofday(&time_end, NULL);
      time_partic+=time_spent();
      
      // allocate analyser
      gettimeofday(&time_start, NULL);
      engine->init(particles);
      gettimeofday(&time_end, NULL);
      time_alloc+=time_spent();

      // analyse stable cones
      gettimeofday(&time_start, NULL);
      nb_proto = engine->get_stable_cones(R);
      gettimeofday(&time_end, NULL);
      
      nb_total+=engine->nb_tot;
      //nb_tested+=engine->hc->n_cones; //cones.size();
      nb_stables+=nb_proto;
      time_analyse+=time_spent();

      // split&merge
      gettimeofday(&time_start, NULL);
      sm->init(particles, &(engine->protocones), R);
      nb_jets+=sm->perform(SMT);
      gettimeofday(&time_end, NULL);
      time_splitmerge+=time_spent();      

      fprintf(fnum, "%d\t%d\n", N, nb_proto);
    }

    fprintf(flux, "%d\t%le\t%le\t%le\t%le\n", N, 
	    time_partic/(1.0*Nruns), time_alloc/(1.0*Nruns), time_analyse/(1.0*Nruns), time_splitmerge/(1.0*Nruns));
    fprintf(fnum2, "%d\t%le\t%le\t%le\t%le\n", N, 
	    nb_total/(1.0*Nruns), nb_tested/(1.0*Nruns), nb_stables/(1.0*Nruns), nb_jets/(1.0*Nruns));

    N+=part_inc;
    if (N==(part_inc<<3))
      part_inc <<= 1;
    //  } while (N<=1024);
  } while (N<=512);
  
  // free engine
  delete engine;
  delete sm;      

  fclose(flux);
  fclose(fnum);
  fclose(fnum2);
  fprintf(stdout, "\n");

  cout << "bye..." << endl;

  return 0;
}
